import pathlib
import sys

import numpy as np
import pytest
import tidynamics

sys.path.insert(0, f"{pathlib.Path(__file__).parents[1].resolve().as_posix()}/src")
from mdcraft.algorithm import correlation  # noqa: E402

# Generate random times series
rng = np.random.default_rng()
shape = (*rng.integers(2, 50, size=3), 3)
vec_ones = np.ones(shape)
vec_random_1 = rng.random(shape)
vec_random_2 = rng.random(shape)

# Calculate reference ACFs and CCFs for the random time series using
# tidynamics
acf_scalar = tidynamics.acf(vec_random_1[0, :, 0, 0])
acf_multi_scalar = np.stack(
    [tidynamics.acf(v) for v in vec_random_1[0, :, :, 0].T]
).T
acf_block_scalar = np.stack(
    [tidynamics.acf(v) for v in vec_random_1[:, :, 0, 0]]
)
acf_vector = tidynamics.acf(vec_random_1[0, :, 0])
acf_multi_vector = np.stack(
    [tidynamics.acf(v) for v in np.swapaxes(vec_random_1[0], 0, 1)]
).T
acf_block_vector = np.stack(
    [tidynamics.acf(v) for v in vec_random_1[:, :, 0]]
)
ccf_scalar = tidynamics.correlation(vec_random_1[0, :, 0, 0],
                                    vec_random_2[0, :, 0, 0])
ccf_multi_scalar = np.stack(
    [tidynamics.correlation(v1, v2)
     for v1, v2 in zip(vec_random_1[0, :, :, 0].T,
                       vec_random_2[0, :, :, 0].T)]
).T
ccf_block_scalar = np.stack(
    [tidynamics.correlation(v1, v2)
     for v1, v2 in zip(vec_random_1[:, :, 0, 0],
                       vec_random_2[:, :, 0, 0])]
)
ccf_vector = tidynamics.correlation(vec_random_1[0, :, 0],
                                    vec_random_2[0, :, 0])
ccf_multi_vector = np.stack(
    [tidynamics.correlation(v1, v2)
     for v1, v2 in zip(np.swapaxes(vec_random_1[0], 0, 1),
                       np.swapaxes(vec_random_2[0], 0, 1))]
).T
ccf_block_vector = np.stack(
    [tidynamics.correlation(v1, v2)
     for v1, v2 in zip(vec_random_1[:, :, 0],
                       vec_random_2[:, :, 0])]
)

def test_func_correlation_fft():

    # TEST CASE 1: ACF of empty 1D array; raises ValueError
    with pytest.raises(ValueError):
        correlation.correlation_fft(np.empty(0))

    # TEST CASE 2: ACF of empty 2D array; raises ValueError
    with pytest.raises(ValueError):
        correlation.correlation_fft(np.empty((0, 3)))

    # TEST CASE 3: ACF of 5D array; raises ValueError
    with pytest.raises(ValueError):
        correlation.correlation_fft(np.empty((2, 2, 2, 2, 2)))

    # TEST CASE 4: ACF of 3D array along invalid axis; raises ValueError
    with pytest.raises(ValueError):
        correlation.correlation_fft(np.empty((2, 2, 2)), axis=2)

    # TEST CASE 5: ACF of time series of ones
    assert np.allclose(correlation.correlation_fft(vec_ones[0, :, 0, 0]), 1)

    # TEST CASE 6: ACF of time series of ones for multiple entities
    with pytest.warns(UserWarning):
        assert np.allclose(
            correlation.correlation_fft(vec_ones[0, :, :, 0]),
            1
        )

    # TEST CASE 7: ACF of blocked time series of ones
    with pytest.warns(UserWarning):
        assert np.allclose(
            correlation.correlation_fft(vec_ones[:, :, 0, 0]),
            1
        )

    # TEST CASE 8: ACF of blocked time series of ones for multiple entities
    with pytest.warns(UserWarning):
        assert np.allclose(
            correlation.correlation_fft(vec_ones[:, :, :, 0]),
            1
        )

    # TEST CASE 9: ACF of time series of 1-vectors
    with pytest.warns(UserWarning):
        assert np.allclose(
            correlation.correlation_fft(vec_ones[0, :, 0], vector=True),
            shape[-1]
        )

    # TEST CASE 10: ACF of time series of 1-vectors for multiple entities
    with pytest.warns(UserWarning):
        assert np.allclose(
            correlation.correlation_fft(vec_ones[0], vector=True),
            shape[-1]
        )

    # TEST CASE 11: ACF of blocked time series of 1-vectors
    with pytest.warns(UserWarning):
        assert np.allclose(
            correlation.correlation_fft(vec_ones[:, :, 0], vector=True),
            shape[-1]
        )

    # TEST CASE 12: ACF of blocked time series of 1-vectors for multiple
    # entities
    assert np.allclose(correlation.correlation_fft(vec_ones, vector=True),
                       shape[-1])

    # TEST CASE 13: ACF of time series of random scalars
    assert np.allclose(correlation.correlation_fft(vec_random_1[0, :, 0, 0]),
                       acf_scalar)

    # TEST CASE 14: ACF of time series of random scalars for multiple
    # entities
    acf = correlation.correlation_fft(vec_random_1[0, :, :, 0], axis=0)
    assert np.allclose(acf, acf_multi_scalar)
    assert np.allclose(
        correlation.correlation_fft(vec_random_1[0, :, :, 0], average=True,
                                    axis=0),
        acf.mean(axis=1)
    )

    # TEST CASE 15: ACF of blocked time series of random scalars
    assert np.allclose(
        correlation.correlation_fft(vec_random_1[:, :, 0, 0], axis=1),
        acf_block_scalar
    )

    # TEST CASE 16: ACF of blocked time series of random scalars for
    # multiple entities
    acf = correlation.correlation_fft(vec_random_1[:, :, :, 0], axis=1)
    assert np.allclose(acf[0], acf_multi_scalar)
    assert np.allclose(acf[:, :, 0], acf_block_scalar)
    assert np.allclose(
        correlation.correlation_fft(vec_random_1[:, :, :, 0], average=True,
                                    axis=1),
        acf.mean(axis=2)
    )

    # TEST CASE 17: ACF of time series of random vectors
    assert np.allclose(
        correlation.correlation_fft(vec_random_1[0, :, 0], axis=0, vector=True),
        acf_vector
    )

    # TEST CASE 18: ACF of time series of random vectors for multiple
    # entities
    assert np.allclose(
        correlation.correlation_fft(vec_random_1[0], axis=0, vector=True),
        acf_multi_vector
    )

    # TEST CASE 19: ACF of blocked time series of random vectors
    assert np.allclose(
        correlation.correlation_fft(vec_random_1[:, :, 0], axis=1, vector=True),
        acf_block_vector
    )

    # TEST CASE 20: ACF of blocked time series of random vectors for
    # multiple entities
    acf = correlation.correlation_fft(vec_random_1, vector=True)
    assert np.allclose(acf[0], acf_multi_vector)
    assert np.allclose(acf[:, :, 0], acf_block_vector)

    # TEST CASE 21: CCF of empty 1D arrays
    with pytest.raises(ValueError):
        correlation.correlation_fft(np.empty(0), np.empty(0))

    # TEST CASE 22: CCF of empty 2D array
    with pytest.raises(ValueError):
        correlation.correlation_fft(np.empty((0, 3)), np.empty((0, 3)))

    # TEST CASE 23: CCF of 5D arrays
    with pytest.raises(ValueError):
        correlation.correlation_fft(np.empty((2, 2, 2, 2, 2)),
                                    np.empty((2, 2, 2, 2, 2)))

    # TEST CASE 24: CCF of asymmetric 2D arrays
    with pytest.raises(ValueError):
        correlation.correlation_fft(np.empty((2, 3)), np.empty((3, 2)))

    # TEST CASE 25: CCF of 3D arrays along invalid axis
    with pytest.raises(ValueError):
        correlation.correlation_fft(np.empty((2, 2, 2)), np.empty((2, 2, 2)),
                                    axis=2)

    # TEST CASE 26: CCF of time series of random scalars
    assert np.allclose(
        correlation.correlation_fft(vec_random_1[0, :, 0, 0].tolist(),
                                    vec_random_2[0, :, 0, 0].tolist()),
        ccf_scalar
    )

    # TEST CASE 27: CCF of time series of random scalars for multiple
    # entities
    assert np.allclose(
        correlation.correlation_fft(vec_random_1[0, :, :, 0],
                                    vec_random_2[0, :, :, 0], axis=0),
        ccf_multi_scalar
    )

    # TEST CASE 28: CCF of blocked time series of random scalars
    assert np.allclose(
        correlation.correlation_fft(vec_random_1[:, :, 0, 0],
                                    vec_random_2[:, :, 0, 0], axis=1),
        ccf_block_scalar
    )

    # TEST CASE 29: CCF of blocked time series of random scalars for
    # multiple entities
    ccf = correlation.correlation_fft(vec_random_1[:, :, :, 0],
                                      vec_random_2[:, :, :, 0], axis=1)
    assert np.allclose(ccf[0], ccf_multi_scalar)
    assert np.allclose(ccf[:, :, 0], ccf_block_scalar)

    # TEST CASE 30: CCF of time series of random vectors
    assert np.allclose(
        correlation.correlation_fft(vec_random_1[0, :, 0],
                                    vec_random_2[0, :, 0],
                                    axis=0, vector=True),
        ccf_vector
    )

    # TEST CASE 31: CCF of time series of random vectors for multiple
    # entities
    assert np.allclose(
        correlation.correlation_fft(vec_random_1[0], vec_random_2[0],
                                    axis=0, vector=True),
        ccf_multi_vector
    )

    # TEST CASE 32: CCF of blocked time series of random vectors
    assert np.allclose(
        correlation.correlation_fft(vec_random_1[:, :, 0],
                                    vec_random_2[:, :, 0],
                                    axis=1, vector=True),
        ccf_block_vector
    )

    # TEST CASE 33: CCF of blocked time series of random vectors for
    # multiple entities
    ccf = correlation.correlation_fft(vec_random_1, vec_random_2,
                                      vector=True)
    assert np.allclose(ccf[0], ccf_multi_vector)
    assert np.allclose(ccf[:, :, 0], ccf_block_vector)

def test_func_correlation_shift():

    # TEST CASE 1: ACF of empty 1D array
    with pytest.raises(ValueError):
        correlation.correlation_shift(np.empty(0))

    # TEST CASE 2: ACF of empty 2D array
    with pytest.raises(ValueError):
        correlation.correlation_shift(np.empty((0, 3)))

    # TEST CASE 3: ACF of 5D array
    with pytest.raises(ValueError):
        correlation.correlation_shift(np.empty((2, 2, 2, 2, 2)))

    # TEST CASE 4: ACF of 3D array along invalid axis
    with pytest.raises(ValueError):
        correlation.correlation_shift(np.empty((2, 2, 2)), axis=2)

    # TEST CASE 5: ACF of time series of random scalars
    assert np.allclose(correlation.correlation_shift(vec_random_1[0, :, 0, 0]),
                       acf_scalar)

    # TEST CASE 6: ACF of time series of random scalars for multiple
    # entities
    with pytest.warns(UserWarning):
        acf = correlation.correlation_shift(vec_random_1[0, :, :, 0])
    assert np.allclose(acf, acf_multi_scalar)
    assert np.allclose(
        correlation.correlation_shift(vec_random_1[0, :, :, 0],
                                      average=True, axis=0),
        acf.mean(axis=1)
    )

    # TEST CASE 7: ACF of blocked time series of random scalars
    assert np.allclose(
        correlation.correlation_shift(vec_random_1[:, :, 0, 0], axis=1),
        acf_block_scalar
    )

    # TEST CASE 8: ACF of blocked time series of random scalars for
    # multiple entities
    acf = correlation.correlation_shift(vec_random_1[:, :, :, 0], axis=1)
    assert np.allclose(acf[0], acf_multi_scalar)
    assert np.allclose(acf[:, :, 0], acf_block_scalar)
    assert np.allclose(
        correlation.correlation_shift(vec_random_1[:, :, :, 0],
                                      average=True, axis=1),
        acf.mean(axis=2)
    )

    # TEST CASE 9: ACF of time series of random vectors
    assert np.allclose(
        correlation.correlation_shift(vec_random_1[0, :, 0], axis=0,
                                      vector=True),
        acf_vector
    )

    # TEST CASE 10: ACF of time series of random vectors for multiple
    # entities
    assert np.allclose(
        correlation.correlation_shift(vec_random_1[0], axis=0,
                                      vector=True),
        acf_multi_vector
    )

    # TEST CASE 11: ACF of blocked time series of random vectors
    assert np.allclose(
        correlation.correlation_shift(vec_random_1[:, :, 0], axis=1,
                                      vector=True),
        acf_block_vector
    )

    # TEST CASE 12: ACF of blocked time series of random vectors for
    # multiple entities
    acf = correlation.correlation_shift(vec_random_1, vector=True)
    assert np.allclose(acf[0], acf_multi_vector)
    assert np.allclose(acf[:, :, 0], acf_block_vector)

    # TEST CASE 13: CCF of empty 1D arrays
    with pytest.raises(ValueError):
        correlation.correlation_shift(np.empty(0), np.empty(0))

    # TEST CASE 14: CCF of empty 2D array
    with pytest.raises(ValueError):
        correlation.correlation_shift(np.empty((0, 3)), np.empty((0, 3)))

    # TEST CASE 15: CCF of 5D arrays
    with pytest.raises(ValueError):
        correlation.correlation_shift(np.empty((2, 2, 2, 2, 2)),
                                      np.empty((2, 2, 2, 2, 2)))

    # TEST CASE 16: CCF of asymmetric 2D arrays
    with pytest.raises(ValueError):
        correlation.correlation_shift(np.empty((2, 3)), np.empty((3, 2)))

    # TEST CASE 17: CCF of 3D arrays along invalid axis
    with pytest.raises(ValueError):
        correlation.correlation_shift(np.empty((2, 2, 2)), np.empty((2, 2, 2)),
                                      axis=2)

    # TEST CASE 18: CCF of time series of random scalars
    assert np.allclose(
        correlation.correlation_shift(vec_random_1[0, :, 0, 0].tolist(),
                                      vec_random_2[0, :, 0, 0].tolist()),
        ccf_scalar
    )

    # TEST CASE 19: CCF of time series of random scalars for multiple
    # entities
    assert np.allclose(
        correlation.correlation_shift(vec_random_1[0, :, :, 0],
                                      vec_random_2[0, :, :, 0], axis=0),
        ccf_multi_scalar
    )

    # TEST CASE 20: CCF of blocked time series of random scalars
    assert np.allclose(
        correlation.correlation_shift(vec_random_1[:, :, 0, 0],
                                      vec_random_2[:, :, 0, 0], axis=1),
        ccf_block_scalar
    )

    # TEST CASE 21: CCF of blocked time series of random scalars for
    # multiple entities
    ccf = correlation.correlation_shift(vec_random_1[:, :, :, 0],
                                        vec_random_2[:, :, :, 0], axis=1)
    assert np.allclose(ccf[0], ccf_multi_scalar)
    assert np.allclose(ccf[:, :, 0], ccf_block_scalar)

    # TEST CASE 22: CCF of time series of random vectors
    assert np.allclose(
        correlation.correlation_shift(vec_random_1[0, :, 0],
                                      vec_random_2[0, :, 0],
                                      axis=0, vector=True),
        ccf_vector
    )

    # TEST CASE 23: CCF of time series of random vectors for multiple
    # entities
    assert np.allclose(
        correlation.correlation_shift(vec_random_1[0], vec_random_2[0],
                                      axis=0, vector=True),
        ccf_multi_vector
    )

    # TEST CASE 24: CCF of blocked time series of random vectors
    assert np.allclose(
        correlation.correlation_shift(vec_random_1[:, :, 0],
                                      vec_random_2[:, :, 0],
                                      axis=1, vector=True),
        ccf_block_vector
    )

    # TEST CASE 25: CCF of blocked time series of random vectors for
    # multiple entities
    ccf = correlation.correlation_shift(vec_random_1, vec_random_2,
                                        vector=True)
    assert np.allclose(ccf[0], ccf_multi_vector)
    assert np.allclose(ccf[:, :, 0], ccf_block_vector)

    # TEST CASE 26: Doubled ACF for time series of random scalars
    assert np.allclose(
        correlation.correlation_shift(vec_random_1[0, :, 0, 0], double=True),
        correlation.correlation_shift(vec_random_1[0, :, 0, 0], double=True)
    )

    # TEST CASE 27: Overlapped CCF for time series of random scalars
    assert np.allclose(
        correlation.correlation_shift(vec_random_1[0, :, 0, 0],
                                      vec_random_2[0, :, 0, 0],
                                      axis=0, double=True),
        correlation.correlation_fft(vec_random_1[0, :, 0, 0],
                                    vec_random_2[0, :, 0, 0],
                                    axis=0, double=True)
    )

    # TEST CASE 28: Overlapped CCF for blocked time series of random scalars
    assert np.allclose(
        correlation.correlation_shift(vec_random_1[:, :, 0, 0],
                                      vec_random_2[:, :, 0, 0],
                                      axis=1, double=True),
        correlation.correlation_fft(vec_random_1[:, :, 0, 0],
                                    vec_random_2[:, :, 0, 0],
                                    axis=1, double=True)
    )

# Generate simple trajectories and calculate their MSDs and CD
traj_1 = np.array(((0, 0, 0), (1, 1, 1), (2, 2, 2), (3, 3, 3)))
msd_1 = np.einsum("td,td->t", traj_1, traj_1)
traj_2 = np.array(((0, 1, 2), (2, 3, 4), (4, 5, 6), (6, 7, 8)))
msd_2 = ((traj_2 - traj_2[0]) ** 2).sum(axis=1)
cd = (traj_1 * (traj_2 - traj_2[0])).sum(axis=1)

def test_func_msd_fft():

    # TEST CASE 1: MSD of empty trajectory
    with pytest.raises(ValueError):
        correlation.msd_fft(np.empty(0))

    # TEST CASE 2: MSD of 5D trajectory
    with pytest.raises(ValueError):
        correlation.msd_fft(np.empty((2, 2, 2, 2, 2)))

    # TEST CASE 3: CD of asymmetric trajectories
    with pytest.raises(ValueError):
        correlation.msd_fft(traj_1, traj_2[:1])

    # TEST CASE 4: MSD of 3D array along invalid axis
    with pytest.raises(ValueError):
        correlation.msd_fft(np.empty((2, 2, 2)), axis=2)

    # TEST CASE 5: MSD of first simple trajectory
    assert np.allclose(correlation.msd_fft(traj_1.tolist()), msd_1)

    # TEST CASE 6: MSD of second simple trajectory
    assert np.allclose(correlation.msd_fft(traj_2), msd_2)

    # TEST CASE 7: CD of simple trajectories
    assert np.allclose(correlation.msd_fft(traj_1, traj_2.tolist()), cd)

    # TEST CASE 8: MSD of first simple trajectory replicated for multiple
    # particles
    with pytest.warns(UserWarning):
        assert np.allclose(
            correlation.msd_fft(
                np.tile(traj_1[:, None], (1, 2, 1)),
                average=False
            )[:, 0],
            msd_1
        )

    # TEST CASE 9: MSD of first simple trajectory replicated for multiple
    # blocks and particles
    assert np.allclose(
        correlation.msd_fft(
            np.tile(traj_1[None, :, None], (2, 1, 2, 1)),
            average=False
        )[0, :, 0],
        msd_1
    )

    # TEST CASE 10: CD of simple trajectories replicated for multiple
    # blocks and particles
    assert np.allclose(
        correlation.msd_fft(
            np.tile(traj_1[None, :, None], (2, 1, 2, 1)),
            np.tile(traj_2[None, :, None], (2, 1, 2, 1))
        )[0],
        cd
    )

def test_func_msd_shift():

    # TEST CASE 1: MSD of empty trajectory
    with pytest.raises(ValueError):
        correlation.msd_shift(np.empty(0))

    # TEST CASE 2: MSD of 5D trajectory
    with pytest.raises(ValueError):
        correlation.msd_shift(np.empty((2, 2, 2, 2, 2)))

    # TEST CASE 3: CD of asymmetric trajectories
    with pytest.raises(ValueError):
        correlation.msd_shift(traj_1, traj_2[:1])

    # TEST CASE 4: MSD of 3D array along invalid axis
    with pytest.raises(ValueError):
        correlation.msd_shift(np.empty((2, 2, 2)), axis=2)

    # TEST CASE 5: MSD of first simple trajectory
    assert np.allclose(correlation.msd_shift(traj_1.tolist()), msd_1)

    # TEST CASE 6: MSD of second simple trajectory
    assert np.allclose(correlation.msd_shift(traj_2), msd_2)

    # TEST CASE 7: CD of simple trajectories
    assert np.allclose(correlation.msd_shift(traj_1, traj_2.tolist()), cd)

    # TEST CASE 8: MSD of first simple trajectory replicated for multiple
    # particles
    with pytest.warns(UserWarning):
        assert np.allclose(
            correlation.msd_shift(
                np.tile(traj_1[:, None], (1, 2, 1)),
                average=False
            )[:, 0],
            msd_1
        )

    # TEST CASE 9: MSD of first simple trajectory replicated for multiple
    # blocks and particles
    assert np.allclose(
        correlation.msd_shift(
            np.tile(traj_1[None, :, None], (2, 1, 2, 1)),
            average=False
        )[0, :, 0],
        msd_1
    )

    # TEST CASE 10: CD of simple trajectories replicated for multiple
    # blocks and particles
    assert np.allclose(
        correlation.msd_shift(
            np.tile(traj_1[None, :, None], (2, 1, 2, 1)),
            np.tile(traj_2[None, :, None], (2, 1, 2, 1))
        )[0],
        cd
    )


"""
Thermodynamic properties
========================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains classes to evaluate thermodynamic properties of
systems, such as the constant-volume heat capacity.
"""

from io import StringIO
from pathlib import Path
from typing import Union
import warnings

import numpy as np
import pandas as pd

from .. import FOUND_OPENMM, Q_, ureg
from ..algorithm.unit import strip_unit
from .base import Hash

if FOUND_OPENMM:
    from openmm import unit


class ConstantVolumeHeatCapacity:
    """
    A serial implementation to caluclate the constant-volume heat
    capacity :math:`C_V` for a canonical (:math:`NVT`) system.

    The constant-volume heat capacity is defined as

    .. math::

       C_V=\\frac{\\langle U^2\\rangle-\\langle U\\rangle^2}
       {k_\\mathrm{B}T^2}

    where :math:`U` is the total potential energy of the system,
    :math:`\\langle\\cdot\\rangle` denotes the ensemble average,
    :math:`k_\\mathrm{B}` is the Boltzmann constant, and :math:`T` is
    the system temperature.

    Parameters
    ----------
    log_file : `str` or `pathlib.Path`, optional
        Log file generated by the simulation. If not provided, the
        potential energies must be provided directly in `energies`.

    log_format : `str`, optional
        Format of the log file. If not provided, the format will be
        determined automatically (if possible).

        **Valid values**: :code:`"lammps"`, :code:`"openmm"`.

    energies : `numpy.ndarray` or `pint.Quantity`, optional
        Potential or total energies :math:`U`. If not provided, the
        log file must be provided in `log_file`.

    temperature : `float`, `openmm.unit.Quantity`, or `pint.Quantity`, \
    optional
        System temperature :math:`T`. If not provided, the averaged
        temperature from the log file (if available) is used.

    reduced : `bool`, keyword-only, default: :code:`False`
        Specifies whether the data is in reduced units.

    sep : `str`, keyword-only, default: :code:`","`
        Delimiter for OpenMM log files.

    Attributes
    ----------
    results.units : `dict`
        Reference units for the results. For example, to get the
        reference units for `results.temperature`, call
        :code:`results.units["temperature"]`.

    results.energies : `numpy.ndarray`
        Potential or total energies :math:`U`.

        **Reference unit**: :math:`\\mathrm{kcal/mol}` or
        :math:`\\mathrm{kJ/mol}`.

    results.temperature : `float`
        System temperature :math:`T`.

        **Reference unit**: :math:`\\mathrm{K}`.

    results.heat_capacity : `float`
        Constant-volume heat capacity :math:`C_V`.

        **Reference unit**: :math:`\\mathrm{kcal/K}` or
        :math:`\\mathrm{kJ/K}`.
        
    Example
    --------
    First, this analysis class must be imported:

    >>> from mdcraft.analysis.thermodynamics import ConstantVolumeHeatCapacity

    Then, after loading the log files (or providing the energies and temperature directly), the class can be instantiated:
    
    >>> cv = ConstantVolumeHeatCapacity("log.lammps)
    >>> cv.run()

    The results can be obtained under the `results` attribute:
    
    >>> cv.results.energies
    >>> cv.results.heat_capacity
    
    """

    _COLUMNS = {
        "lammps": {
            "energy": [
                "TotEng",
                "KinEng",
                "PotEng",
                "E_angle",
                "E_bond",
                "E_coul",
                "E_dihed",
                "E_impro",
                "E_long",
                "E_vdwl",
            ],
            "temperature": "Temp",
        },
        "openmm": {
            "energy": [
                "Total Energy (kJ/mole)",
                "Kinetic Energy (kJ/mole)",
                "Potential Energy (kJ/mole)",
            ],
            "temperature": "Temperature (K)",
        },
    }

    def __init__(
        self,
        log_file: Union[str, Path] = None,
        log_format: str = None,
        *,
        energies: Union[np.ndarray[float], "unit.Quantity", Q_] = None,
        temperature: Union[float, "unit.Quantity", Q_] = None,
        reduced: bool = False,
        sep: str = ",",
    ) -> None:

        self.results = Hash(units={})
        self._reduced = reduced

        if energies:
            self.results.units["energies"] = ureg.kilojoule / ureg.mole
            self.results.energies = strip_unit(energies, self.results.units["energy"])[
                0
            ]
        elif log_file:
            self._file = log_file if isinstance(log_file, Path) else Path(log_file)
            with open(self._file, "r") as f:
                log = f.read()

            # Determine the simulation toolkit used to write the log
            # file
            if log_format is None:
                for f, cs in self._COLUMNS.items():
                    if any(c in log for c in cs["energy"]):
                        log_format = f
                        break
                else:
                    raise ValueError("Could not determine log file format.")
            self._format = log_format

            # Pre-process LAMMPS log files
            if self._format == "lammps":
                if "minimize" in log:
                    log = log[log.index("Minimization stats:") :]
                log = log.split("\n")
                valid = False
                for i, line in enumerate(log):
                    if "Step" in line:
                        log = log[i:]
                        valid = True
                        break
                if not valid:
                    emsg = "No thermodynamic data found in log file " f"'{log_file}'."
                    raise ValueError(emsg)
                log = "\n".join(log)
                log = log[: log.index("Loop time of ")]
                kwargs = {"sep": r"\s+"}
                self.results.units["energies"] = ureg.kilocalorie / ureg.mole
                self.results.units["heat_capacity"] = ureg.kilocalorie / ureg.kelvin
            elif self._format == "openmm":
                kwargs = {"sep": sep}
                if reduced:
                    warnings.warn("OpenMM simulations always use real units.")
                self.results.units["energies"] = ureg.kilojoule / ureg.mole
                self.results.units["heat_capacity"] = ureg.kilojoule / ureg.kelvin

            # Determine which columns to keep.
            if self._COLUMNS[self._format]["energy"][0] in log:
                cols = self._COLUMNS[self._format]["energy"][:1]
            elif self._COLUMNS[self._format]["energy"][1] in log:
                cols = self._COLUMNS[self._format]["energy"][1:2]
                if self._COLUMNS[self._format]["energy"][2] in log:
                    cols.append(self._COLUMNS[self._format]["energy"][2])
                elif any(e in log for e in self._COLUMNS[self._format]["energy"][3:]):
                    for e in self._COLUMNS[self._format]["energy"][3:]:
                        if e in log:
                            cols.append(e)
                else:
                    raise ValueError("Potential energy column not found.")
            else:
                raise ValueError("Total or kinetic energy column not found.")

            df = pd.read_csv(StringIO(log), **kwargs)
            self.results.energies = df[cols].sum(axis=1).to_numpy()
        else:
            raise ValueError("No log file or energy values provided.")

        if temperature is not None:
            self.temperature, self.results.units["temperature"] = strip_unit(
                temperature
            )
            if self.results.units["temperature"] is None:
                self.results.units["temperature"] = ureg.kelvin
        elif log_file is None:
            raise ValueError("No log file or temperature value provided.")
        else:
            self.temperature = df[self._COLUMNS[self._format]["temperature"]].mean()
            self.results.units["temperature"] = ureg.kelvin

    def run(
        self,
        start: int = None,
        stop: int = None,
        step: int = None,
        frames: Union[np.ndarray[int], slice] = None,
    ) -> None:
        """
        Performs the calculation.

        Parameters
        ----------
        start : `int`, optional
            Starting frame for analysis.

        stop : `int`, optional
            Ending frame for analysis.

        step : `int`, optional
            Number of frames to skip between each analyzed frame.

        frames : `slice` or array-like, optional
            Index or logical array of the desired trajectory frames.

        verbose : `bool`, optional
            Determines whether detailed progress is shown.

        **kwargs
            Additional keyword arguments to pass to
            :class:`MDAnalysis.lib.log.ProgressBar`.

        Returns
        -------
        self : `ConstantVolumeHeatCapacity`
            Analysis object with results.
        """

        if frames is None:
            frames = np.arange(start or 0, stop or len(self.results.energies), step)

        U = self.results.energies[frames]
        if self._reduced:
            self.results.heat_capacity = (
                (U**2).mean() - U.mean() ** 2
            ) / self.temperature**2
        else:
            U *= self.results.units["energies"]
            self.results.heat_capacity = strip_unit(
                ((U**2).mean() - U.mean() ** 2)
                / (
                    ureg.avogadro_constant**2
                    * ureg.boltzmann_constant
                    * (self.temperature * self.results.units["temperature"]) ** 2
                ),
                self.results.units["heat_capacity"],
            )[0]

        return self



import glob
import os
import pathlib
import sys
import urllib

import MDAnalysis as mda
from MDAnalysis.analysis.msd import EinsteinMSD
from MDAnalysis.tests.datafiles import RANDOM_WALK, RANDOM_WALK_TOPO
import numpy as np
from scipy.stats import linregress

sys.path.insert(0, f"{pathlib.Path(__file__).parents[1].resolve().as_posix()}/src")
from mdcraft.analysis import transport # noqa: E402

def test_class_onsager_msd():

    universe = mda.Universe(RANDOM_WALK_TOPO, RANDOM_WALK)

    start = 20
    stop = 60
    time = np.arange(universe.trajectory.n_frames)
    msd = EinsteinMSD(universe).run().results.timeseries / 6
    diff = linregress(time[start:stop], msd[start:stop]).slope

    # TEST CASE 1: MSD and diffusion coefficients of a random walk
    # calculated using the Einstein relation
    universe.dimensions = np.array((np.inf, np.inf, np.inf, 90, 90, 90))
    onsager_shift = transport.Onsager(universe.atoms, fft=False, reduced=True).run()
    onsager_shift.calculate_transport_coefficients(start, stop, scale="linear")
    assert np.allclose(onsager_shift.results.msd_self[0, 0], msd)
    assert np.isclose(diff, onsager_shift.results.D_i[0, 0])

    # TEST CASE 2: MSD and diffusion coefficients of a random walk
    # calculated using the FFT-based algorithm
    universe.dimensions = np.array((np.inf, np.inf, np.inf, 90, 90, 90))
    onsager_fft = transport.Onsager(universe.atoms, reduced=True).run()
    onsager_fft.calculate_transport_coefficients(start, stop, scale="linear")
    assert np.allclose(onsager_fft.results.msd_self[0, 0], msd)
    assert np.isclose(diff, onsager_fft.results.D_i[0, 0])

def test_class_onsager_transport_coefficients():

    """
    The test cases are adapted from the "Mean Squared Displacement —
    :code:`MDAnalysis.analysis.msd`" page from the MDAnalysis User Guide
    (https://docs.mdanalysis.org/stable/documentation_pages/analysis/msd.html)
    and uses data from the paper "Onsager Transport Coefficients and
    Transference Numbers in Polyelectrolyte Solutions and Polymerized
    Ionic Liquids" by Fong et al.
    (https://doi.org/10.1021/acs.macromol.0c02001).
    """

    def acf_fft(x):
        N = len(x)
        f = np.fft.fft(x, n=2 * N)
        return np.fft.ifft(f * f.conj())[:N].real / (N * np.ones(N)
                                                          - np.arange(0, N))

    def msd_fft(r):
        N = len(r)
        D = np.append(np.square(r).sum(axis=1), 0)
        Q = 2 * D.sum()
        S1 = np.zeros(N)
        for m in range(N):
            Q = Q - D[m - 1] - D[N - m]
            S1[m] = Q / (N - m)
        return S1 - 2 * sum(acf_fft(r[:, i]) for i in range(r.shape[1]))

    def ccf_fft(x, y):
        N = len(x)
        return np.fft.ifft(
            np.fft.fft(x, n=2 ** (2 * N - 1).bit_length())
            * np.fft.fft(y, n=2 ** (2 * N - 1).bit_length()).conj()
        )[:N].real / (N * np.ones(N) - np.arange(0, N))

    def msd_cross_fft(r, k):
        N = len(r)
        D = np.append(np.multiply(r, k).sum(axis=1), 0)
        Q = 2 * D.sum()
        S1 = np.zeros(N)
        for m in range(N):
            Q = Q - D[m - 1] - D[N - m]
            S1[m] = Q / (N - m)
        return S1 - sum(ccf_fft(r[:, i], k[:, i]) for i in range(r.shape[1])) \
               - sum(ccf_fft(k[:, i], r[:, i]) for i in range(k.shape[1]))

    def calc_L_ii_self(positions):
        L_ii_self = np.zeros(positions.shape[0])
        for i in range(positions.shape[1]):
            L_ii_self += msd_fft(positions[:, i, :])
        return L_ii_self

    def calc_L_ii(positions):
        return msd_fft(positions.sum(axis=1))

    def calc_L_ij(cation_positions, anion_positions):
        return msd_cross_fft(cation_positions.sum(axis=1),
                             anion_positions.sum(axis=1))

    def compute_L_ij(anion_positions, cation_positions, volume):
        return np.vstack(
            (
                calc_L_ii(anion_positions),
                calc_L_ij(cation_positions, anion_positions),
                calc_L_ii(cation_positions),
                calc_L_ii_self(anion_positions),
                calc_L_ii_self(cation_positions)
            )
        ) / (6 * volume)

    def fit_data(times, f, start, stop):
        return linregress(times[start:stop], f[start:stop])[0]

    path = os.getcwd()
    if "tests" in path:
        path_split = path.split("/")
        path = "/".join(path_split[:path_split.index("tests") + 1])
    else:
        path += "/tests"
    if not os.path.isdir(f"{path}/data/onsager"):
        os.makedirs(f"{path}/data/onsager")
    os.chdir(f"{path}/data/onsager")

    url = "https://raw.githubusercontent.com/kdfong/transport-coefficients-MSD/master/example-data"
    if not os.path.isfile("system.data"):
        with urllib.request.urlopen(f"{url}/system.data") as r:
            with open("system.data", "w") as f:
                f.write(r.read().decode())
    for i in range(1, 6):
        if not os.path.isfile(f"traj_{i}.dcd"):
            with urllib.request.urlopen(f"{url}/traj_{i}.dcd") as r:
                with open(f"traj_{i}.dcd", "wb") as f:
                    f.write(r.read())

    dt = 50
    start = 40
    fit_start = 2
    fit_stop = 20
    fit_start_self = 20
    fit_stop_self = 50

    universe = mda.Universe("system.data", glob.glob("*.dcd"), format="LAMMPS")
    groups = [universe.select_atoms(f"type {i}") for i in range(1, 3)]
    positions = [np.zeros((universe.trajectory.n_frames - start, g.n_atoms, 3))
                 for g in groups]
    for i, _ in enumerate(universe.trajectory[40:]):
        com = universe.atoms.center_of_mass(wrap=True)
        for g, p in zip(groups, positions):
            p[i] = g.positions - com
    volume = universe.dimensions[:3].prod()
    times = np.arange(0, (universe.trajectory.n_frames - start) * dt, dt,
                      dtype=int)
    msds = compute_L_ij(*positions, volume)

    onsager = transport.Onsager(groups, temperature=1, center=True,
                                center_atom=True, center_wrap=True,
                                reduced=True, dt=dt).run(start=40)
    onsager.calculate_transport_coefficients(fit_start, fit_stop,
                                             start_self=fit_start_self,
                                             stop_self=fit_stop_self,
                                             scale="linear", enforce_linear=False)

    L_ij_array = np.triu(onsager.results.L_ij)
    for i, (msd, L_ij) in enumerate(zip(msds, L_ij_array[L_ij_array != 0])):

        # TEST CASE 1: Cross displacements of polyelectrolyte system
        assert np.allclose(msd, onsager.results.msd_cross[i, 0] / volume,
                           atol=1e-3)

        # TEST CASE 2: Onsager transport coefficients of polyelectrolyte
        # system
        assert np.isclose(fit_data(times, msd, fit_start, fit_stop), L_ij)

    for i, (msd, L_ii_self) in enumerate(zip(msds[3:], onsager.results.L_ii_self[0])):

        # TEST CASE 1: MSDs of polyelectrolyte system
        assert np.allclose(
            msd, groups[i].n_atoms * onsager.results.msd_self[i, 0] / volume,
            atol=1e-6
        )

        # TEST CASE 2: Self Onsager transport coefficients of
        # polyelectrolyte system
        assert np.isclose(
            fit_data(times, msd, fit_start_self, fit_stop_self),
            L_ii_self
        )


"""
Matplotlib rcParams
===================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module provides optimized Matplotlib rcParams for various
scientific publications.
"""

import shutil

import matplotlib as mpl

# Define figure size guidelines for various publications in inches
FIGURE_SIZE_LIMITS = {
    "acs": {"max_single_width": 3.25, "max_double_width": 7, "max_length": 9.5},
    "aip": {
        "max_single_width": 3.37,
        "max_double_width": 6.69,
        "max_length": 8.25,
        "min_font_size": 8,
    },
    "rsc": {
        "max_single_width": 3.26771654,
        "max_double_width": 6.73228346,
        "max_length": 9.17322835,
    },
}


def update(
    journal: str = None, font_scaling: float = 1, size_scaling: float = 1, **kwargs
) -> None:
    """
    Updates the Matplotlib rcParams at runtime. By default, this
    function overwrites the following settings:

    .. code::

       {
         "axes.labelsize": 9,
         "figure.autolayout": True,
         "font.size": 9,
         "legend.columnspacing": 1,
         "legend.edgecolor": "1",
         "legend.fontsize": 9,
         "legend.handlelength": 1.25,
         "legend.labelspacing": 0.25,
         "savefig.dpi": 1_200,
         "xtick.labelsize": 9,
         "ytick.labelsize": 9,
         "text.usetex": True # If LaTeX is available
       }

    If a supported journal acronym is provided as the first argument,
    the default figure size will also be updated.

    Parameters
    ----------
    journal : `str`, optional
        Journal acronym used to update the default figure size.

        .. container::

           **Valid values**:

           * :code:`"acs"`: American Chemical Society.
           * :code:`"aip"`: American Institute of Physics.
           * :code:`"rsc"`: Royal Society of Chemistry.

    font_scaling : `float`, optional
        Scaling factor to apply to the default font size. This
        value is multiplied by the default font size to obtain the
        final font size.

    size_scaling : `float`, optional
        Scaling factor to apply to the default figure size. This
        value is multiplied by the default figure size to obtain the
        final figure size.

    **kwargs
        Additional rcParams to update passed to
        :meth:`matplotlib.rcParams.update`.
    """

    fig_size = (
        {}
        if journal is None
        else {
            "figure.figsize": (
                size_scaling * FIGURE_SIZE_LIMITS[journal]["max_single_width"],
                size_scaling * 3 * FIGURE_SIZE_LIMITS[journal]["max_single_width"] / 4,
            )
        }
    )

    mpl.rcParams.update(
        {
            "axes.labelsize": font_scaling * 9,
            "figure.autolayout": True,
            "font.size": font_scaling * 9,
            "legend.columnspacing": 1,
            "legend.edgecolor": "1",
            "legend.fontsize": font_scaling * 9,
            "legend.handlelength": 1.25,
            "legend.labelspacing": 0.25,
            "savefig.dpi": 1_200,
            "xtick.labelsize": font_scaling * 9,
            "ytick.labelsize": font_scaling * 9,
            "text.usetex": bool(shutil.which("latex")),
        }
        | fig_size
        | kwargs
    )



"""
Axis components
===============
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module provides additional functionality for Matplotlib axes.
"""

from typing import Any

import matplotlib as mpl
import numpy as np


def set_up_tabular_legend(
    rows: list[str],
    cols: list[str],
    *,
    hlabel: str = None,
    vlabel: str = None,
    hla: str = "left",
    vla: str = "top",
    condense: bool = False,
    **kwargs,
) -> tuple[dict[str, Any], int, int]:
    r"""
    Sets up a tabular legend for a :class:`matplotlib.axes.Axes` object.

    Parameters
    ----------
    rows : `tuple` or `list`
        Raw string representations of the row values.

    cols : `tuple` or `list`
        Raw string representations of the column values.

    hlabel : `str`, keyword-only, optional
        Horizontal label for column values.

    vlabel : `str`, keyword-only, optional
        Vertical label for row values.

    hla : `str`, keyword-only, default: :code:`"left"`
        Alignment for `hlabel`.

        .. container::

           **Valid values**:

           * :code:`"left"`: Left-aligned text.
           * :code:`"center"`: Horizontally centered text.

    vla : `str`, keyword-only, default: :code:`"top"`
        Alignment for `vlabel`.

        .. container::

           **Valid values**:

           * :code:`"top"`: Top-aligned text.
           * :code:`"center"`: Vertically centered text.

    condense : `bool`, keyword-only, default: :code:`False`
        Condenses the legend by placing `vlabel` in the empty top-left
        corner. Cannot be used when no `vlabel` is specified or in
        conjuction with :code:`vla="center"` (which will take priority).

    **kwargs :
        Keyword arguments passed to :meth:`matplotlib.axes.Axes.legend`.

    Returns
    -------
    properties : `dict`
        Properties of the tabular legend to be unpacked and used in the
        :meth:`matplotlib.axes.Axes.legend` call.

        .. container::

           * handles (`list`): :obj:`matplotlib.artist` objects to be
             added to the legend.
           * labels (`list`): Labels to be shown next to the
             :obj:`matplotlib.artist` objects in the legend.
           * ncol (`int`): Number of columns in the legend.
           * kwargs (`dict`): Keyword arguments passed to
             :meth:`matplotlib.axes.Axes.legend`.

    nrow : `int`
        Number of rows in the legend.

    idx_start : `int`
        Index at which to start storing handles for
        :obj:`matplotlib.artist` objects.

    Notes
    -----
    Condensing the legend can cause alignment issues in the first column
    containing the row values if the row values are not of comparable
    length to the rows label. An easy but imprecise fix is to center the
    shorter values using the width of the largest
    :class:`matplotlib.transforms.Bbox` in the first column. Sample code
    for use in your main script utilizing the outputs of this function is
    provided below:

    .. code-block::

       (props, nrow, start) = tabular_legend(..., handletextpad=-5/4)
       fig, ax = plt.subplots(...)
       for i, x in enumerate(...):
           for j, y in enumerate(...):
               props["handles"][start + i * nrow + j], = ax.plot(...)
       ax.set_xlabel(...)
       ax.set_ylabel(...)
       lgd = ax.legend(**props)
       fig.canvas.draw()
       texts = lgd.get_texts()[:nrow]
       bounds = [t.get_window_extent().bounds[2] / 2 for t in texts]
       center = max(bounds)
       for k, t in enumerate(texts):
           t.set_position((center - bounds[k], 0))
       plt.show()
    """

    hpad = bool(vlabel) - condense + 1
    vpad = bool(hlabel) + 1
    nrow = len(rows) + vpad
    ncol = len(cols) + hpad

    labels = ["" for _ in range(nrow * ncol)]
    if vlabel:
        labels[vpad + (len(rows) // 2 if vla == "center" else -condense)] = vlabel
    iv = vpad + nrow * (bool(vlabel) - condense)
    labels[iv : iv + len(rows)] = rows
    if hlabel:
        labels[(2 + (hla == "center") * (int(np.ceil(len(cols) / 2)) - 1)) * nrow] = (
            hlabel
        )
    labels[hpad * nrow + bool(hlabel) :: nrow] = cols

    return (
        {
            "handles": [
                mpl.patches.Rectangle((0, 0), 0.1, 0.1, ec="none", fill=False)
                for _ in range(len(labels))
            ],
            "labels": labels,
            "ncol": ncol,
            **kwargs,
        },
        nrow,
        iv + nrow,
    )



"""
Topology transformations
========================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains implementations of common LAMMPS topology
transformations, like the generation of initial particle positions.
"""

from io import TextIOWrapper
from numbers import Real
from typing import Any, Union

import numpy as np

from .. import FOUND_OPENMM, Q_, ureg
from ..algorithm import topology as topo

if FOUND_OPENMM:
    from openmm import app, unit


def create_atoms(
    dimensions: Union[np.ndarray[float], "unit.Quantity", Q_, "app.Topology"],
    N: int = None,
    N_p: int = 1,
    *,
    lattice: str = None,
    length: Union[float, "unit.Quantity"] = 0.34,
    flexible: bool = False,
    bonds: bool = False,
    angles: bool = False,
    dihedrals: bool = False,
    randomize: bool = False,
    length_unit: Union["unit.Unit", ureg.Unit] = None,
    wrap: bool = False,
) -> Any:
    """
    Generates initial particle positions for coarse-grained simulations.

    .. seealso::

       This function is an alias for
       :func:`mdcraft.algorithm.topology.create_atoms`.
    """

    return topo.create_atoms(
        dimensions,
        N,
        N_p,
        lattice=lattice,
        length=length,
        flexible=flexible,
        bonds=bonds,
        angles=angles,
        dihedrals=dihedrals,
        randomize=randomize,
        length_unit=length_unit,
        wrap=wrap,
    )


def write_data(
    file: Union[str, TextIOWrapper],
    positions: tuple[np.ndarray[float]],
    *,
    bonds: tuple[np.ndarray[int]] = None,
    angles: tuple[np.ndarray[int]] = None,
    dihedrals: tuple[np.ndarray[int]] = None,
    impropers: tuple[np.ndarray[int]] = None,
    dimensions: np.ndarray[float] = None,
    tilt: np.ndarray[float] = None,
    charges: np.ndarray[float] = None,
    masses: np.ndarray[float] = None,
) -> None:
    r"""
    Writes topological data to a LAMMPS data file in :code:`atom_style full`.

    Parameters
    ----------
    file : `str` or `_io.TextIOWrapper`
        LAMMPS data file.

    positions : `tuple`
        Atomic positions. Each element of the tuple should contain
        atoms of the same atom type.

        **Shape**: Tuple of arrays with shape :math:`(*,\,3)`.

        **Reference units**: :math:`\mathrm{Å}`.

    bonds : `tuple`, keyword-only, optional
        Pairs of indices of bonded atoms. Each element of the tuple
        should contain bonds of the same bond type.

        **Shape**: Tuple of arrays with shape :math:`(*,\,2)`.

    angles : `tuple`, keyword-only, optional
        Triples of indices of atoms that form an angle. Each element of
        the tuple should contain angles of the same angle type.

        **Shape**: Tuple of arrays with shape :math:`(*,\,3)`.

    dihedrals : `tuple`, keyword-only, optional
        Quadruples of indices of atoms that form a dihedral. Each
        element of the tuple should contain dihedrals of the same
        dihedral type.

        **Shape**: Tuple of arrays with shape :math:`(*,\,4)`.

    impropers : `tuple`, keyword-only, optional
        Quadruples of indices of atoms that form an improper. Each
        element of the tuple should contain impropers of the same
        improper type.

        **Shape**: Tuple of arrays with shape :math:`(*,\,4)`.

    dimensions : array-like, keyword-only, optional
        Box dimensions. If three values are provided, the box
        dimensions are assumed to be from :math:`0` to the specified
        values. If six values are provided, the box dimensions go from
        the three values in the first column to the three values in the
        second column.

        **Shape**: :math:`(3,)` or :math:`(3,\,2)`.

        **Reference units**: :math:`\mathrm{Å}`.

    tilt : array-like, keyword-only, optional
        Box :math:`xy`, :math:`xz`, and :math:`yz` tilt factors.

        **Shape**: :math:`(3,)`.

    charges : array-like, keyword-only, optional
        Atomic charges.

        **Shape**: :math:`(N,)`.

    masses : array-like, keyword-only, optional
        Atomic masses.

        **Shape**: :math:`(N,)`.
    """

    if isinstance(file, str):
        file = open(file, "w")

    # Write header
    file.write("LAMMPS Description\n\n")
    n_atoms_type = [len(p) for p in positions]
    n_atoms = sum(n_atoms_type)
    file.write(f"{n_atoms} atoms\n")
    file.write(f"{len(positions)} atom types\n")
    if bonds is not None:
        n_bonds_type = [len(b) for b in bonds]
        file.write(f"{sum(n_bonds_type)} bonds\n")
        file.write(f"{len(bonds)} bond types\n")
    if angles is not None:
        n_angles_type = [len(a) for a in angles]
        file.write(f"{sum(n_angles_type)} angles\n")
        file.write(f"{len(angles)} angle types\n")
    if dihedrals is not None:
        n_dihedrals_type = [len(d) for d in dihedrals]
        file.write(f"{sum(n_dihedrals_type)} dihedrals\n")
        file.write(f"{len(dihedrals)} dihedral types\n")
    if impropers is not None:
        n_impropers_type = [len(i) for i in impropers]
        file.write(f"{sum(n_impropers_type)} impropers\n")
        file.write(f"{len(impropers)} improper types\n")
    if dimensions is not None:
        if dimensions.ndim == 1:
            dimensions = np.vstack((np.zeros(3), dimensions)).T
        for i, d in enumerate(dimensions):
            a = chr(120 + i)
            file.write(f"{d[0]:.6g} {d[1]:.6g} {a}lo {a}hi\n")
    if tilt is not None:
        file.write(f"{tilt[0]:.6g} {tilt[1]:.6g} {tilt[2]:.6g} xy xz yz\n")

    # Write masses
    if masses is not None:
        if len(masses) != len(positions):
            emsg = "Number of masses must match number of atom types."
            raise ValueError(emsg)
        file.write("\nMasses\n\n")
        for i, m in enumerate(masses):
            file.write(f"{i + 1} {m:.6g}\n")

    # Write atom positions
    if charges is None:
        charges = np.zeros(n_atoms)
    if len(charges) == len(positions):
        charges = list(charges)
        for i, (qs, n) in enumerate(zip(charges, n_atoms_type)):
            if isinstance(qs, Real):
                charges[i] *= np.ones(n)
    elif len(charges) == n_atoms:
        charges = np.array_split(charges, np.cumsum(n_atoms)[:-1])
    else:
        raise ValueError("'charges' has an invalid shape.")
    file.write("\nAtoms # full\n\n")
    for t, (pos, qs) in enumerate(zip(positions, charges)):
        start = sum(n_atoms_type[:t])
        for i, (p, q) in enumerate(zip(pos, qs)):
            file.write(
                f"{start + i + 1} {start + i + 1} {t + 1} {q:.6g} "
                f"{p[0]:.6g} {p[1]:.6g} {p[2]:.6g}\n"
            )

    # Write bonds
    if bonds is not None:
        file.write("\nBonds\n\n")
        for t, b in enumerate(bonds):
            start = sum(n_bonds_type[:t])
            for i, (a, b) in enumerate(b):
                file.write(f"{start + i + 1} {t + 1} {a} {b}\n")

    # Write angles
    if angles is not None:
        file.write("\nAngles\n\n")
        for t, a in enumerate(angles):
            start = sum(n_angles_type[:t])
            for i, (a, b, c) in enumerate(a):
                file.write(f"{start + i + 1} {t + 1} {a} {b} {c}\n")

    # Write dihedrals
    if dihedrals is not None:
        file.write("\nDihedrals\n\n")
        for t, d in enumerate(dihedrals):
            start = sum(n_dihedrals_type[:t])
            for i, (a, b, c, d) in enumerate(d):
                file.write(f"{start + i + 1} {t + 1} {a} {b} {c} {d}\n")

    # Write impropers
    if impropers is not None:
        file.write("\nImpropers\n\n")
        for t, i in enumerate(impropers):
            start = sum(n_impropers_type[:t])
            for j, (a, b, c, d) in enumerate(i):
                file.write(f"{start + j + 1} {t + 1} {a} {b} {c} {d}\n")

    file.close()



"""
LAMMPS tools
============
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module provides a number of extensions to the LAMMPS simulation
toolkit.
"""

from . import topology

__all__ = ["topology"]



"""
Plotting settings and tools
===========================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains Matplotlib rcParams and tools for generating
aesthetically-pleasing scientific figures.
"""

from . import axis, color, rcparam

__all__ = ["axis", "color", "rcparam"]



"""
Polymeric properties
====================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains classes to determine the structural
and dynamical properties of polymeric systems.
"""

from typing import Union
import warnings

import MDAnalysis as mda
from MDAnalysis.lib.log import ProgressBar
import numba
import numpy as np
from scipy import optimize, special

from . import structure
from .base import Hash, DynamicAnalysisBase, NumbaAnalysisBase
from .. import FOUND_OPENMM, Q_, ureg
from ..algorithm import accelerated, correlation
from ..algorithm.molecule import center_of_mass, radius_of_gyration
from ..algorithm.topology import unwrap, unwrap_edge
from ..algorithm.unit import strip_unit
from ..algorithm.utility import get_closest_factors
from ..fit.exponential import stretched_exp

if FOUND_OPENMM:
    from openmm import unit


class _PolymerAnalysisBase(DynamicAnalysisBase):
    r"""
    An analysis base object for polymer systems.

    Parameters
    ----------
    groups : `MDAnalysis.AtomGroup` or array-like
        Groups of polymers to be analyzed.

        .. note::

           All polymers in each group must have the same chain length.

    groupings : `str` or array-like, default: :code:`"atoms"`
        Determines whether the centers of mass are used in lieu of
        individual atom positions. If `groupings` is a `str`, the same
        value is used for all `groups`.

        .. note::

           In a standard trajectory file, segments (or chains) contain
           residues (or molecules), and residues contain atoms. This
           heirarchy must be adhered to for this analysis module to
           function correctly. If your trajectory file does not contain
           the correct residue or segment information, provide the
           number of chains and chain lengths in `n_chains` and
           `n_monomers`, respectively.

        .. container::

           **Valid values**:

           * :code:`"atoms"`: Atom positions (generally or for
             coarse-grained simulations).
           * :code:`"residues"`: Residues' centers of mass (for
             atomistic simulations).

    n_chains : `int` or array-like, optional
        Number of chains :math:`M` in each polymer group. Must be
        provided if the trajectory does not adhere to the standard
        container heirarchy. If an `int` is provided, the same value is
        used for all groups.

        **Shape**: :math:`(N_\mathrm{groups},)`.

    n_monomers : `int` or array-like, optional
        Number of monomers :math:`N` in each chain in each polymer
        group. Must be provided if the trajectory does not adhere to the
        standard container heirarchy. If an `int` is provided, the same
        value is used for all groups.

        **Shape**: :math:`(N_\mathrm{groups},)`.

    unwrap : `bool`, keyword-only, default: :code:`False`
        Determines whether atom positions are unwrapped.

    parallel : `bool`, keyword-only, default: :code:`False`
        Determines whether the analysis is performed in parallel.

    verbose : `bool`, keyword-only, default: :code:`True`
        Determines whether detailed progress is shown.

    **kwargs
        Additional keyword arguments to pass to
        :class:`MDAnalysis.analysis.base.AnalysisBase`.

    Attributes
    ----------
    universe : `MDAnalysis.Universe`
        :class:`MDAnalysis.core.universe.Universe` object containing all
        information describing the system.
    """

    def __init__(
        self,
        groups: Union[mda.AtomGroup, tuple[mda.AtomGroup]],
        groupings: Union[str, tuple[str]] = "atoms",
        n_chains: Union[int, tuple[int]] = None,
        n_monomers: Union[int, tuple[int]] = None,
        *,
        unwrap: bool = False,
        parallel: bool = False,
        verbose: bool = True,
        **kwargs,
    ) -> None:

        self._groups = [groups] if isinstance(groups, mda.AtomGroup) else groups
        self.universe = self._groups[0].universe
        super().__init__(self.universe.trajectory, parallel, verbose, **kwargs)

        self._dimensions = self.universe.dimensions
        if self._dimensions is not None:
            self._dimensions = self._dimensions[:3].copy()

        GROUPINGS = {"atoms", "residues"}
        self._n_groups = len(self._groups)
        if isinstance(groupings, str):
            if groupings not in GROUPINGS:
                emsg = (
                    f"Invalid grouping '{groupings}'. Valid values: "
                    "'" + "', '".join(GROUPINGS) + "'."
                )
                raise ValueError(emsg)
            self._groupings = self._n_groups * [groupings]
        else:
            if self._n_groups != len(groupings):
                emsg = (
                    "The shape of 'groupings' is incompatible with " "that of 'groups'."
                )
                raise ValueError(emsg)
            for gr in groupings:
                if gr not in GROUPINGS:
                    emsg = (
                        f"Invalid grouping '{gr}'. Valid "
                        "values: '" + "', '".join(GROUPINGS) + "'."
                    )
                    raise ValueError(emsg)
            self._groupings = groupings

        if n_chains is None or n_monomers is None:
            self._internal = True
            self._n_chains = np.empty(self._n_groups, dtype=int)
            self._n_monomers = np.empty_like(self._n_chains)
            for i, ag in enumerate(self._groups):
                self._n_chains[i] = ag.segments.n_segments
                self._n_monomers[i] = ag.n_atoms // self._n_chains[i]
        else:
            self._internal = False
            if isinstance(n_chains, (int, np.integer)):
                self._n_chains = n_chains * np.ones(self._n_groups, dtype=int)
            elif self._n_groups == len(n_chains):
                self._n_chains = n_chains
            else:
                emsg = (
                    "The shape of 'n_chains' is incompatible with " "that of 'groups'."
                )
                raise ValueError(emsg)
            if isinstance(n_monomers, (int, np.integer)):
                self._n_monomers = n_monomers * np.ones(n_monomers, dtype=int)
            elif self._n_groups == len(n_monomers):
                self._n_monomers = n_monomers
            else:
                emsg = (
                    "The shape of 'n_monomers' is incompatible "
                    "with that of 'groups'."
                )
                raise ValueError(emsg)

        self._unwrap = unwrap
        self._verbose = verbose


class Gyradius(_PolymerAnalysisBase):
    """
    Serial and parallel implementations to calculate the radius of
    gyration :math:`R_\\mathrm{g}` of a polymer.

    The radius of gyration is used to describe the dimensions of a
    polymer chain, and is defined as

    .. math::

        R_\\mathrm{g}=\\sqrt{
        \\frac{\\sum_i^N m_i\\|\\mathbf{r}_i
        -\\mathbf{R}_\\mathrm{com}\\|^2}{\\sum_i^N m_i}}

    where :math:`m_i` and :math:`\\mathbf{r}_i` are the mass and
    position, respectively, of particle :math:`i`, and
    :math:`\\mathbf{R}_\\mathrm{com}` is the center of mass.

    Parameters
    ----------
    groups : `MDAnalysis.AtomGroup` or array-like
        Groups of polymers to be analyzed.

        .. note::

           All polymers in each group must have the same chain length.

    groupings : `str` or array-like, default: :code:`"atoms"`
        Determines whether the centers of mass are used in lieu of
        individual atom positions. If `groupings` is a `str`, the same
        value is used for all `groups`.

        .. note::

           In a standard trajectory file, segments (or chains) contain
           residues (or molecules), and residues contain atoms. This
           heirarchy must be adhered to for this analysis module to
           function correctly. If your trajectory file does not contain
           the correct residue or segment information, provide the
           number of chains and chain lengths in `n_chains` and
           `n_monomers`, respectively.

        .. container::

           **Valid values**:

           * :code:`"atoms"`: Atom positions (generally or for
             coarse-grained simulations).
           * :code:`"residues"`: Residues' centers of mass (for
             atomistic simulations).

    n_chains : `int` or array-like, optional
        Number of chains :math:`M` in each polymer group. Must be
        provided if the trajectory does not adhere to the standard
        container heirarchy. If an `int` is provided, the same value is
        used for all groups.

        **Shape**: :math:`(N_\\mathrm{groups},)`.

    n_monomers : `int` or array-like, optional
        Number of monomers :math:`N` in each chain in each polymer
        group. Must be provided if the trajectory does not adhere to the
        standard container heirarchy. If an `int` is provided, the same
        value is used for all groups.

        **Shape**: :math:`(N_\\mathrm{groups},)`.

    components : `bool`, keyword-only, default: :code:`False`
        Specifies whether the components of the radii of gyration are
        calculated and returned instead.

    unwrap : `bool`, keyword-only, default: :code:`False`
        Determines whether atom positions are unwrapped.

    parallel : `bool`, keyword-only, default: :code:`False`
        Determines whether the analysis is performed in parallel.

    verbose : `bool`, keyword-only, default: :code:`True`
        Determines whether detailed progress is shown.

    **kwargs
        Additional keyword arguments to pass to
        :class:`MDAnalysis.analysis.base.AnalysisBase`.

    Attributes
    ----------
    universe : `MDAnalysis.Universe`
        :class:`MDAnalysis.core.universe.Universe` object containing all
        information describing the system.

    results.units : `dict`
        Reference units for the results. For example, to get the
        reference units for `results.gyradii`, call
        :code:`results.units["gyradii"]`.

    results.gyradii : `numpy.ndarray`
        Radii of gyration.

        **Shape**: :math:`(N_\\mathrm{groups},\\,N_\\mathrm{frames})` or
        :math:`(N_\\mathrm{groups},\\,N_\\mathrm{frames},\\,3)`.

        **Reference unit**: :math:`\\mathrm{Å}`.

    Example
    --------
    First, this analysis class must be imported:

    >>> from mdcraft.analysis.polymer import Gyradius

    Then, after loading a simulation trajectory:

    >>> universe = mda.Universe("simulation.nc", "topology.cif")

    We must then select the atom-groups to be analyzed:

    >>> ag = universe.select_atoms("resname POLY")

    The gyradius class can then be built and run, remembering to specify the number of chains and monomers per chain:

    >>> gyradius = Gyradius(ag, n_chains=10, n_monomers=100)
    >>> gyradius.run()

    The results can be obtained under the `results` attribute:

    >>> gyradius.results.gyradii

    """

    def __init__(
        self,
        groups: Union[mda.AtomGroup, tuple[mda.AtomGroup]],
        groupings: Union[str, tuple[str]] = "atoms",
        n_chains: Union[int, tuple[int]] = None,
        n_monomers: Union[int, tuple[int]] = None,
        *,
        components: bool = False,
        unwrap: bool = False,
        parallel: bool = False,
        verbose: bool = True,
        **kwargs,
    ):

        super().__init__(
            groups,
            groupings,
            n_chains,
            n_monomers,
            unwrap=unwrap,
            parallel=parallel,
            verbose=verbose,
            **kwargs,
        )

        if self.universe.dimensions is None and unwrap:
            emsg = (
                "System dimensions were not found, but are "
                "required when 'unwrap=True'."
            )
            raise ValueError(emsg)

        self._Ns_p = self._n_chains * self._n_monomers
        self._N_p = self._Ns_p.sum()
        self._slices = []
        _ = 0
        for N_p in self._Ns_p:
            self._slices.append(slice(_, _ + N_p))
            _ += N_p

        self._components = components

    def _prepare(self) -> None:

        if self._unwrap:

            # Navigate to first frame in analysis
            ts = self._sliced_trajectory[0]

            # Get system dimensions
            dimensions = ts.dimensions[:3]

            # Preallocate arrays to determine the number of periodic
            # boundary crossings for each entity
            self._positions_old = np.empty((self._N_p, 3))
            for ag, gr, s, M, N in zip(
                self._groups,
                self._groupings,
                self._slices,
                self._n_chains,
                self._n_monomers,
            ):
                if self._internal and gr == "residues":
                    ag.unwrap()
                    self._positions_old[s] = center_of_mass(ag, gr)
                else:
                    positions = unwrap_edge(
                        positions=ag.positions,
                        bonds=np.array(
                            [
                                (i * N + j, i * N + j + 1)
                                for i in range(M)
                                for j in range(N - 1)
                            ]
                        ),
                        dimensions=dimensions,
                        masses=ag.masses,
                    )
                    self._positions_old[s] = (
                        positions
                        if gr == "atoms"
                        else center_of_mass(
                            positions=positions.reshape(M, N, -1, 3),
                            masses=ag.masses.reshape(M, N, -1),
                        )
                    )
            self._images = np.zeros((self._N_p, 3), dtype=int)

            # Store unwrapped entity positions in a shared memory array
            # for parallel analysis
            if self._parallel:
                self._positions = np.empty((self.n_frames, self._N_p, 3))
                for i, _ in enumerate(self._sliced_trajectory):
                    for ag, gr, s, M, N in zip(
                        self._groups,
                        self._groupings,
                        self._slices,
                        self._n_chains,
                        self._n_monomers,
                    ):
                        if self._internal and gr == "residues":
                            self._positions[i, s] = center_of_mass(ag, gr)
                        else:
                            self._positions[i, s] = (
                                ag.positions
                                if gr == "atoms"
                                else center_of_mass(
                                    positions=ag.positions.reshape(M, N, -1, 3),
                                    masses=ag.masses.reshape(M, N, -1),
                                )
                            )

                    # Globally unwrap entity positions for correct
                    # centers of mass and radii of gyration
                    unwrap(
                        self._positions[i],
                        self._positions_old,
                        dimensions,
                        thresholds=dimensions / 2,
                        images=self._images,
                    )

        # Preallocate arrays to hold radii of gyration
        if not self._parallel:
            shape = [self._n_groups, self.n_frames]
            if self._components:
                shape.append(3)
            self.results.gyradii = np.empty(shape)

        # Store reference units
        self.results.units = Hash({"gyradii": ureg.angstrom})

    def _single_frame(self) -> None:

        for i, (ag, gr, M, N, s) in enumerate(
            zip(
                self._groups,
                self._groupings,
                self._n_chains,
                self._n_monomers,
                self._slices,
            )
        ):

            # Store atom or center-of-mass positions in the current frame
            if self._internal and gr == "residues":
                positions = center_of_mass(ag, gr)
            else:
                positions = (
                    ag.positions
                    if gr == "atoms"
                    else center_of_mass(
                        positions=ag.positions.reshape(M, N, -1, 3),
                        masses=ag.masses.reshape(M, N, -1),
                    )
                )

            # Globally unwrap entity positions for correct centers of
            # mass and radii of gyration
            if self._unwrap:
                dimensions = self._ts.dimensions[:3]
                unwrap(
                    positions,
                    self._positions_old[s],
                    dimensions,
                    thresholds=dimensions / 2,
                    images=self._images[s],
                )

            # Compute the radii of gyration
            self.results.gyradii[i, self._frame_index] = radius_of_gyration(
                grouping="segments",
                positions=positions.reshape((M, N, 3)),
                masses=ag.masses.reshape((M, N)),
                components=self._components,
            ).mean(axis=0)

    def _single_frame_parallel(self, index: int) -> tuple[int, np.ndarray[float]]:

        # Set current trajectory frame
        self._sliced_trajectory[index]

        # Preallocate array to store radii of gyration
        shape = [self._n_groups]
        if self._components:
            shape.append(3)
        gyradii = np.empty(shape)

        for i, (ag, gr, M, N, s) in enumerate(
            zip(
                self._groups,
                self._groupings,
                self._n_chains,
                self._n_monomers,
                self._slices,
            )
        ):

            # Calculate or get entity positions
            if self._unwrap:
                positions = self._positions[index, s]
            elif self._internal and gr == "residues":
                positions = center_of_mass(ag, gr)
            else:
                positions = (
                    ag.positions
                    if gr == "atoms"
                    else center_of_mass(
                        positions=ag.positions.reshape(M, N, -1, 3),
                        masses=ag.masses.reshape(M, N, -1),
                    )
                )

            # Compute the radii of gyration
            gyradii[i] = radius_of_gyration(
                grouping="segments",
                positions=positions.reshape((M, N, 3)),
                masses=ag.masses.reshape((M, N)),
                components=self._components,
            ).mean(axis=0)

        return index, gyradii

    def _conclude(self) -> None:

        # Consolidate parallel results and clean up memory by deleting
        # arrays that will not be reused
        if self._parallel:
            self._results = sorted(self._results)
            self.results.gyradii = np.stack([r[1] for r in self._results], axis=1)

            del self._results
            if self._unwrap:
                del self._positions
        if self._unwrap:
            del self._positions_old, self._images


def correlation_fft(
    x: np.ndarray[Union[float, complex]],
    y: np.ndarray[Union[float, complex]] = None,
    /,
    axis: int = None,
    *,
    average: bool = False,
    double: bool = False,
    vector: bool = False,
) -> np.ndarray[Union[float, complex]]:
    r"""
    Evaluates the autocorrelation functions (ACF)
    :math:`\mathrm{R_\mathbf{XX}}(\tau)` or cross-correlation functions
    (CCF) :math:`\mathrm{R_\mathbf{XY}}(\tau)` of time series
    :math:`\mathbf{X}(t)` and :math:`\mathbf{Y}(t)` using fast Fourier
    transforms (FFT).

    .. seealso::

       This function is an alias for
       :func:`mdcraft.algorithm.correlation.correlation_fft`.
    """

    return correlation.correlation_fft(
        x, y, axis, average=average, double=double, vector=vector
    )


def correlation_shift(
    x: np.ndarray[Union[float, complex]],
    y: np.ndarray[Union[float, complex]] = None,
    /,
    axis: int = None,
    *,
    average: bool = False,
    double: bool = False,
    vector: bool = False,
) -> np.ndarray[Union[float, complex]]:
    r"""
    Evaluates the autocorrelation functions (ACF)
    :math:`\mathrm{R_\mathbf{XX}}(\tau)` or cross-correlation functions
    (CCF) :math:`\mathrm{R_\mathbf{XY}}(\tau)` of time series
    :math:`\mathbf{X}(t)` and :math:`\mathbf{Y}(t)` directly by using
    sliding windows.

    .. seealso::

       This function is an alias for
       :func:`mdcraft.algorithm.correlation.correlation_shift`.
    """

    return correlation.correlation_shift(
        x, y, axis, average=average, double=double, vector=vector
    )


def calculate_relaxation_time(
    times: np.ndarray[float], acf: np.ndarray[float]
) -> float:
    r"""
    Calculates the orientational relaxation time :math:`\tau_\mathrm{r}`
    of polymers using end-to-end vector autocorrelation function (ACF)
    time series :math:`C_\mathrm{ee}(t)`.

    A stretched exponential function with :math:`\tau` and :math:`\beta`
    as coefficients,

    .. math::

       C_\mathrm{ee}(t)=\exp{\left[-(t/\tau)^\beta\right]}

    is fitted to the ACF time series, and the relaxation time is
    estimated using the first moment of :math:`C_\mathrm{ee}`,

    .. math::

       \tau_\mathrm{r}=\int_0^\infty C_\mathrm{ee}(t)\,dt
       =\frac{\tau}{\beta}\Gamma\left(\frac{1}{\beta}\right)

    where :math:`\Gamma` is the gamma function.

    Parameters
    ----------
    times : `numpy.ndarray`
        Changes in time :math:`t-t_0`.

        **Shape**: :math:`(N_\mathrm{frames},)`.

        **Reference unit**: :math:`\mathrm{ps}`.

    acf : `numpy.ndarray`
        End-to-end vector ACF :math:`C_\mathrm{ee}(t)`.

        **Shape**:
        :math:`(N_\mathrm{groups},\,N_\mathrm{bins},\,N_\mathrm{frames})`.

    Returns
    -------
    relaxation_time : `float`
        Average orientational relaxation time :math:`\tau_\mathrm{r}`.

        **Reference unit**: :math:`\mathrm{ps}`.
    """

    tau_r, beta = optimize.curve_fit(
        stretched_exp, times / times[1], acf, bounds=(0, np.inf)
    )[0]
    return tau_r * times[1] * special.gamma((beta + 1) / beta)


class EndToEndVector(_PolymerAnalysisBase):
    """
    A serial implementation to calculate the end-to-end vector
    autocorrelation function (ACF) :math:`C_\\mathrm{ee}(t)` and
    estimate the orientational relaxation time :math:`\\tau_\\mathrm{r}`
    of a polymer.

    The end-to-end vector ACF is defined as

    .. math::

       C_\\mathrm{ee}(t)=\\frac{\\langle\\mathbf{R}_\\mathrm{ee}(t)
       \\cdot\\mathbf{R}_\\mathrm{ee}(0)\\rangle}
       {\\langle\\mathbf{R}_\\mathrm{ee}^2\\rangle}

    where :math:`\\mathbf{R}_\\mathrm{ee}=\\mathbf{r}_N-\\mathbf{r}_1`
    is the end-to-end vector.

    The orientational relaxation time can then be estimated by fitting
    a stretched exponential function

    .. math::

       C_\\mathrm{ee}=\\exp{\\left[-(t/\\tau)^\\beta\\right]}

    to the end-to-end vector ACF and evaluating

    .. math::

       \\tau_\\mathrm{r}=\\int_0^\\infty C_\\mathrm{ee}\\,dt
       =\\frac{\\tau}{\\beta}\\Gamma\\left(\\frac{1}{\\beta}\\right)

    Parameters
    ----------
    groups : `MDAnalysis.AtomGroup` or array-like
        Groups of polymers to be analyzed.

        .. note::

           All polymers in each group must have the same chain length.

    groupings : `str` or array-like, default: :code:`"atoms"`
        Determines whether the centers of mass are used in lieu of
        individual atom positions. If `groupings` is a `str`, the same
        value is used for all `groups`.

        .. note::

           In a standard trajectory file, segments (or chains) contain
           residues (or molecules), and residues contain atoms. This
           heirarchy must be adhered to for this analysis module to
           function correctly. If your trajectory file does not contain
           the correct residue or segment information, provide the
           number of chains and chain lengths in `n_chains` and
           `n_monomers`, respectively.

        .. container::

           **Valid values**:

           * :code:`"atoms"`: Atom positions (generally or for
             coarse-grained simulations).
           * :code:`"residues"`: Residues' centers of mass (for
             atomistic simulations).

    n_chains : `int` or array-like, optional
        Number of chains :math:`M` in each polymer group. Must be
        provided if the trajectory does not adhere to the standard
        container heirarchy. If an `int` is provided, the same value is
        used for all groups.

        **Shape**: :math:`(N_\\mathrm{groups},)`.

    n_monomers : `int` or array-like, optional
        Number of monomers :math:`N` in each chain in each polymer
        group. Must be provided if the trajectory does not adhere to the
        standard container heirarchy. If an `int` is provided, the same
        value is used for all groups.

        **Shape**: :math:`(N_\\mathrm{groups},)`.

    n_blocks : `int`, keyword-only, default: :code:`1`
        Number of blocks to split the trajectory into.

    dt : `float`, `openmm.unit.Quantity`, or `pint.Quantity`, \
    keyword-only, optional
        Time between frames :math:`\\Delta t`. While this is normally
        determined from the trajectory, the trajectory may not have the
        correct information if the data is in reduced units. For
        example, if the reduced timestep is :math:`0.01` and trajectory
        data was outputted every :math:`10,000` timesteps, then
        :math:`\\Delta t=100`.

        **Reference unit**: :math:`\\mathrm{ps}`.

    fft : `bool`, keyword-only, default: :code:`True`
        Determines whether fast Fourier transforms (FFT) are used to
        evaluate the ACFs.

    unwrap : `bool`, keyword-only, default: :code:`False`
        Determines whether atom positions are unwrapped.

    verbose : `bool`, keyword-only, default: :code:`True`
        Determines whether detailed progress is shown.

    **kwargs
        Additional keyword arguments to pass to
        :class:`MDAnalysis.analysis.base.AnalysisBase`.

    Attributes
    ----------
    universe : `MDAnalysis.Universe`
        :class:`MDAnalysis.core.universe.Universe` object containing all
        information describing the system.

    results.units : `dict`
        Reference units for the results. For example, to get the
        reference units for `results.times`, call
        :code:`results.units["times"]`.

    results.times : `numpy.ndarray`
        Changes in time :math:`t-t_0`.

        **Shape**: :math:`(N_\\mathrm{frames},)`.

        **Reference unit**: :math:`\\mathrm{ps}`.

    results.acf : `numpy.ndarray`
        End-to-end vector ACFs :math:`C_\\mathrm{ee}(t)`.

        **Shape**:
        :math:`(N_\\mathrm{groups},\\,N_\\mathrm{blocks},\\,N_\\mathrm{frames})`.

    results.relaxation_times : `numpy.ndarray`
        Average orientational relaxation times :math:`\\tau_\\mathrm{r}`.

        **Shape**: :math:`(N_\\mathrm{groups},\\,N_\\mathrm{blocks})`.

        **Reference units**: :math:`\\mathrm{ps}`.
    
    Example
    --------
    First, this analysis class must be imported:

    >>> from mdcraft.analysis.polymer import EndToEndVector

    Then, after loading a simulation trajectory:

    >>> universe = mda.Universe("simulation.nc", "topology.cif")

    We must then select the atom-groups to be analyzed:

    >>> ag = universe.select_atoms("resname POLY")

    The `EndToEndVector` class can then be built and run, remembering to specify the number of chains and monomers per chain:
    
    >>> e2e = EndToEndVector(ag, n_chains=10, n_monomers=100)
    >>> e2e.run()

    The results can be obtained under the `results` attribute:
    
    >>> e2e.results.acf
    
    One can also obtain the relaxation times by calling the following:
    
    >>> e2e.calculate_relaxation_time()
    >>> e2e.results.relaxation_times
    
    These can be saved by doing the following:
    
    >>> e2e.save("e2e")
    
    """

    def __init__(
        self,
        groups: Union[mda.AtomGroup, tuple[mda.AtomGroup]],
        groupings: Union[str, tuple[str]] = "atoms",
        n_chains: Union[int, tuple[int]] = None,
        n_monomers: Union[int, tuple[int]] = None,
        *,
        n_blocks: int = 1,
        dt: Union[float, "unit.Quantity", Q_] = None,
        fft: bool = True,
        unwrap: bool = False,
        verbose: bool = True,
        **kwargs,
    ) -> None:

        # Disable parallel support for this analysis class
        if "parallel" in kwargs:
            wmsg = (
                "The 'EndToEndVector' analysis class does not "
                "support multithreading."
            )
            warnings.warn(wmsg)
            del kwargs["parallel"]

        super().__init__(
            groups,
            groupings,
            n_chains,
            n_monomers,
            unwrap=unwrap,
            verbose=verbose,
            **kwargs,
        )

        self._M = self._n_chains.sum()
        self._slices = []
        _ = 0
        for M in self._n_chains:
            self._slices.append(slice(_, _ + M))
            _ += M

        self._n_blocks = n_blocks
        self._dt = strip_unit(dt, "ps")[0] or self._trajectory.dt
        self._fft = fft

    def _prepare(self) -> None:

        # Ensure frames are evenly spaced and proceed forward in time
        if hasattr(self._sliced_trajectory, "frames"):
            dfs = np.diff(self._sliced_trajectory.frames)
            if (df := dfs[0]) <= 0 or not np.allclose(dfs, df):
                emsg = (
                    "The selected frames must be evenly spaced and "
                    "proceed forward in time."
                )
                raise ValueError(emsg)
        elif (df := self.step) <= 0:
            raise ValueError("The analysis must proceed forward in time.")

        # Determine number of frames used when the trajectory is split
        # into blocks
        self._n_frames_block = self.n_frames // self._n_blocks
        self._n_frames = self._n_blocks * self._n_frames_block
        if (extra_frames := self.n_frames - self._n_frames) > 0:
            wmsg = (
                f"The trajectory is not divisible into {self._n_blocks:,} "
                f"blocks, so the last {extra_frames:,} frame(s) will be "
                "discarded. To maximize performance, set appropriate "
                "starting and ending frames in run() so that the number "
                "of frames to be analyzed is divisible by the number of "
                "blocks."
            )
            warnings.warn(wmsg)

        # Preallocate arrays to store end-to-end vectors
        self._e2e = np.empty((self.n_frames, self._M, 3))

        if self._unwrap:

            # Navigate to first frame in analysis
            self._sliced_trajectory[0]

            # Preallocate arrays to determine the number of periodic
            # boundary crossings for the first and last monomer in each
            # chain
            self._positions_end_old = np.empty((self._M, 2, 3))
            for ag, gr, s, M, N in zip(
                self._groups,
                self._groupings,
                self._slices,
                self._n_chains,
                self._n_monomers,
            ):
                if self._internal and gr == "residues":
                    ag.unwrap()
                    self._positions_end_old[s] = np.stack(
                        [
                            center_of_mass(s.residues[[0, -1]].atoms, "residues")
                            for s in ag.segments
                        ]
                    )
                else:
                    positions = unwrap_edge(
                        positions=ag.positions,
                        bonds=np.array(
                            [
                                (i * N + j, i * N + j + 1)
                                for i in range(M)
                                for j in range(N - 1)
                            ]
                        ),
                        dimensions=self._dimensions,
                        masses=ag.masses,
                    ).reshape(M, N, -1, 3)[:, (0, -1)]
                    self._positions_end_old[s] = (
                        positions[:, :, 0]
                        if gr == "atoms"
                        else center_of_mass(
                            positions=positions,
                            masses=ag.masses.reshape(M, N, -1)[:, (0, -1)],
                        )
                    )
            self._images = np.zeros((self._M, 2, 3), dtype=int)
            self._thresholds = self._dimensions / 2

        # Preallocate array to store end-to-end vector ACFs
        self.results.acf = np.empty(
            (self._n_groups, self._n_blocks, self._n_frames_block)
        )

        # Store time changes
        self.results.times = df * self._dt * np.arange(self._n_frames_block)

        # Store reference units
        self.results.units = {"times": ureg.picosecond}

    def _single_frame(self) -> None:

        for ag, gr, s, M, N in zip(
            self._groups,
            self._groupings,
            self._slices,
            self._n_chains,
            self._n_monomers,
        ):

            # Store ending monomer or center-of-mass positions in the
            # current frame
            if self._internal and gr == "residues":
                positions_end = np.stack(
                    center_of_mass(s.residues[[0, -1]].atoms, "residues")
                    for s in ag.segments
                )
            else:
                positions_end = ag.positions.reshape(M, N, -1, 3)[:, (0, -1)]
                positions_end = (
                    positions_end[:, :, 0]
                    if gr == "atoms"
                    else center_of_mass(
                        positions=positions_end,
                        masses=ag.masses.reshape(M, N, -1)[:, (0, -1)],
                    )
                )

            # Globally unwrap entity positions for correct end-to-end
            # vectors
            if self._unwrap:
                unwrap(
                    positions_end,
                    self._positions_end_old[s],
                    self._dimensions,
                    thresholds=self._thresholds,
                    images=self._images[s],
                )

            # Compute the end-to-end vectors
            self._e2e[self._frame_index, s] = np.diff(positions_end, axis=1)[:, 0]

    def _conclude(self) -> None:

        # Clean up memory by deleting arrays that will not be reused
        if self._unwrap:
            del self._positions_end_old, self._images, self._thresholds

        # Compute end-to-end vector ACFs
        _acf = correlation_fft if self._fft else correlation_shift
        for i, (s, M) in ProgressBar(enumerate(zip(self._slices, self._n_chains))):
            self.results.acf[i] = _acf(
                (
                    self._e2e[:, s]
                    / np.linalg.norm(self._e2e[:, s], axis=-1, keepdims=True)
                ).reshape(self._n_blocks, -1, M, 3),
                average=True,
                vector=True,
            )

    def calculate_relaxation_times(self) -> None:
        """
        Calculates the orientational relaxation times.
        """

        if not hasattr(self.results, "acf"):
            emsg = (
                "Call EndToEndVector.run() before "
                "EndToEndVector.calculate_relaxation_times()."
            )
            raise RuntimeError(emsg)

        self.results.relaxation_times = np.empty((self._n_groups, self._n_blocks))
        self.results.units["relaxation_times"] = ureg.picosecond

        for i, block in enumerate(self.results.acf):
            for j, acf in enumerate(block):
                valid = np.where(acf >= 0)[0]
                self.results.relaxation_times[i, j] = calculate_relaxation_time(
                    self.results.times[valid], acf[valid]
                )


class SingleChainStructureFactor(NumbaAnalysisBase, _PolymerAnalysisBase):
    """
    Serial and parallel implementations to calculate the single-chain
    structure factor :math:`S_\\mathrm{sc}(q)` of a homopolymer.

    It is defined as

    .. math::

       S_{\\mathrm{sc}}(q)=\\frac{1}{MN}\\left\\langle
       \\sum_{m=1}^M\\sum_{j=1}^{N}\\sum_{k=1}^{N}\\exp{
       [i\\mathbf{q}\\cdot(\\mathbf{r}_j-\\mathbf{r}_k)]}\\right\\rangle

    where :math:`M` is the number of chains, :math:`N` is the chain
    length, :math:`\\mathbf{q}` and :math:`q` are the scattering
    wavevector and its magnitude, respectively, and
    :math:`\\mathbf{r}_i` is the position of the :math:`i`-th monomer.

    .. container::

       The single-chain structure factor reveals information about the
       characteristic length scales of the polymer:

       * In the Guinier regime (:math:`qR_g\\ll1`),
         :math:`S_{\\mathrm{sc}}(q)^{-1}\\approx N(1-(qR_g)^2/3)`
         can be used to determine the radius of gyration :math:`R_g`.
       * In the Porod regime (:math:`qR_g\\gg1`),
         :math:`S_{\\mathrm{sc}}(q)=1` since the only contribution is
         the self-scattering of the monomers.
       * In the intermediate regime, the slope :math:`s` of the log-log
         plot of :math:`S_{\\mathrm{sc}}(q)` is related to the scaling
         exponent :math:`\\nu` via :math:`\\nu=-1/s`.

    Parameters
    ----------
    groups : `MDAnalysis.AtomGroup` or array-like
        Groups of polymers to be analyzed.

        .. note::

           All polymers in each group must have the same chain length.

    groupings : `str` or array-like, default: :code:`"atoms"`
        Determines whether the centers of mass are used in lieu of
        individual atom positions. If `groupings` is a `str`, the same
        value is used for all `groups`.

        .. note::

           In a standard trajectory file, segments (or chains) contain
           residues (or molecules), and residues contain atoms. This
           heirarchy must be adhered to for this analysis module to
           function correctly. If your trajectory file does not contain
           the correct residue or segment information, provide the
           number of chains and chain lengths in `n_chains` and
           `n_monomers`, respectively.

        .. container::

           **Valid values**:

           * :code:`"atoms"`: Atom positions (generally or for
             coarse-grained simulations).
           * :code:`"residues"`: Residues' centers of mass (for
             atomistic simulations).

    n_chains : `int` or array-like, optional
        Number of chains :math:`M` in each polymer group. Must be
        provided if the trajectory does not adhere to the standard
        container heirarchy. If an `int` is provided, the same value is
        used for all groups.

        **Shape**: :math:`(N_\\mathrm{groups},)`.

    n_monomers : `int` or array-like, optional
        Number of monomers :math:`N` in each chain in each polymer
        group. Must be provided if the trajectory does not adhere to the
        standard container heirarchy. If an `int` is provided, the same
        value is used for all groups.

        **Shape**: :math:`(N_\\mathrm{groups},)`.

    form : `str`, keyword-only, default: :code:`"exp"`
        Expression used to evaluate the single-chain structure factors.

        .. container::

           **Valid values**:

           * :code:`"exp"`: Exponential form. Slightly faster due to
             fewer mathematical operations.
           * :code:`"trig"`: Trigonometric form. Slightly slower but
             doesn't have overflow issues.

    dimensions : array-like, `openmm.unit.Quantity`, or \
    `pint.Quantity`, keyword-only, optional
        System dimensions :math:`(L_x,\\,L_y,\\,L_z)`. Only used to
        determine the scattering wavevectors when `wavevectors` is not
        specified.

        **Shape**: :math:`(3,)`.

        **Reference unit**: :math:`\\mathrm{Å}`.

    n_points : `int`, keyword-only, default: :code:`32`
        Number of points :math:`n_\\mathrm{points}` in the scattering
        wavevector grid to generate

        .. math::

           \\mathbf{q}=2\\pi\\left(\\frac{a}{L_x},\\,\\frac{b}{L_y},\\,
           \\frac{c}{L_z}\\right)

        where :math:`a`, :math:`b`, and :math:`c` are integers from
        :math:`0` up to :math:`n_\\mathrm{points}-1`.

        Additional wavevectors can be introduced via `n_surfaces` and
        `n_surface_points` for more accurate structure factors at small
        wavenumbers. Alternatively, the desired wavevectors can be
        specified directly in `wavevectors`.

    n_surfaces : `int`, keyword-only, optional
        Number of spherical surfaces in the first octant that intersect
        with the grid wavevectors along the three coordinate axes for
        which to introduce extra wavevectors for more accurate structure
        factor values. Only available if the system is perfectly cubic.

    n_surface_points : `int`, keyword-only, default: :code:`8`
        Number of extra wavevectors to introduce per spherical surface.
        Has no effect if `n_surfaces` is not specified.

    q_max : `float`, `openmm.unit.Quantity`, or `pint.Quantity`, \
    keyword-only, optional
        Maximum wavenumber :math:`q_\\mathrm{max}`.

        **Reference unit**: :math:`\\mathrm{Å}^{-1}`.

    wavevectors : array-like, `openmm.unit.Quantity`, or `pint.Quantity`, \
    keyword-only, optional
        Scattering wavevectors for which to compute structure factors.
        Has precedence over `n_points`, `n_surfaces`, and
        `n_surface_points` if specified.

        **Shape**: :math:`(N_q,\\,3)`.

        **Reference unit**: :math:`\\mathrm{Å}^{-1}`.

    sort : `bool`, keyword-only, default: :code:`True`
        Determines whether the results are sorted by the wavenumbers.

    unique : `bool`, keyword-only, default: :code:`True`
        Determines whether structure factors for the same wavenumber
        are grouped and averaged.

    parallel : `bool`, keyword-only, default: :code:`False`
        Determines whether the analysis is performed in parallel.

    verbose : `bool`, keyword-only, default: :code:`True`
        Determines whether detailed progress is shown.

    **kwargs
        Additional keyword arguments to pass to
        :class:`MDAnalysis.analysis.base.AnalysisBase`.

    Attributes
    ----------
    universe : `MDAnalysis.Universe`
        :class:`MDAnalysis.core.universe.Universe` object containing all
        information describing the system.

    results.units : `dict`
        Reference units for the results. For example, to get the
        reference units for :code:`results.wavenumbers`, call
        :code:`results.units["wavenumbers"]`.

    results.wavenumbers : `numpy.ndarray`
        Wavenumbers :math:`q`.

        **Shape**: :math:`(N_q,)`.

        **Reference unit**: :math:`\\mathrm{Å}^{-1}`.

    results.scsf : `numpy.ndarray`
        Single-chain structure factors :math:`S_\\mathrm{sc}(q)`.

        **Shape**: :math:`(N_q,)`.
        
    Example
    --------
    First, this analysis class must be imported:

    >>> from mdcraft.analysis.polymer import SingleChainStructureFactor

    Then, after loading a simulation trajectory:

    >>> universe = mda.Universe("simulation.nc", "topology.cif")

    We must then select the atom-groups to be analyzed:

    >>> ag = universe.select_atoms("resname POLY")

    The `SingleChainStructureFactor` class can then be built and run, remembering to specify the number of chains and monomers per chain:
    
    >>> scsf = SingleChainStructureFactor(ag, n_chains=10, n_monomers=100)
    >>> scsf.run()

    The results can be obtained under the `results` attribute:
    
    >>> scsf.results.scsf
    
    """

    def __init__(
        self,
        groups: mda.AtomGroup,
        groupings: str = "atoms",
        n_chains: int = None,
        n_monomers: int = None,
        *,
        form: str = "exp",
        dimensions: Union[np.ndarray[float], "unit.Quantity", Q_] = None,
        n_points: int = 32,
        n_surfaces: int = None,
        n_surface_points: int = 8,
        q_max: Union[float, "unit.Quantity", Q_] = None,
        wavevectors: np.ndarray[float] = None,
        sort: bool = True,
        unique: bool = True,
        unwrap: bool = False,
        parallel: bool = False,
        verbose: bool = True,
        **kwargs,
    ) -> None:

        # Specify 'parallel=False' to use the NumbaAnalysisBase.run method
        # instead of the ParallelAnalysisBase.run method
        _PolymerAnalysisBase.__init__(
            self,
            groups,
            groupings,
            n_chains,
            n_monomers,
            unwrap=unwrap,
            parallel=False,
            verbose=verbose,
            **kwargs,
        )

        if wavevectors is not None:
            self._wavevectors = wavevectors
        else:
            dimensions = strip_unit(dimensions or self.universe.dimensions[:3], "Å")[0]
            if dimensions is None:
                emsg = (
                    "System dimensions were not found, but are "
                    "required when 'wavevectors' is not specified."
                )
                raise ValueError(emsg)
            elif len(dimensions) != 3:
                raise ValueError("'dimensions' must have length 3.")
            dimensions = np.asarray(dimensions)

            if np.allclose(dimensions, dimensions[0]):
                grid = 2 * np.pi * np.arange(n_points) / dimensions[0]
                self._wavevectors = np.stack(np.meshgrid(grid, grid, grid), -1).reshape(
                    -1, 3
                )
                if n_surfaces:
                    n_theta, n_phi = get_closest_factors(
                        n_surface_points, 2, reverse=True
                    )
                    theta = np.linspace(
                        np.pi / (2 * n_theta + 4),
                        np.pi / 2 - np.pi / (2 * n_theta + 4),
                        n_theta,
                    )
                    phi = np.linspace(
                        np.pi / (2 * n_phi + 4),
                        np.pi / 2 - np.pi / (2 * n_phi + 4),
                        n_phi,
                    )
                    self._wavevectors = np.vstack(
                        (
                            self._wavevectors,
                            np.einsum(
                                "o,tpd->otpd",
                                grid[1 : n_surfaces + 1],
                                np.stack(
                                    (
                                        np.sin(theta) * np.cos(phi)[:, None],
                                        np.sin(theta) * np.sin(phi)[:, None],
                                        np.tile(np.cos(theta)[None, :], (n_phi, 1)),
                                    ),
                                    axis=-1,
                                ),
                            ).reshape((n_surfaces * n_surface_points, 3)),
                        )
                    )
            else:
                self._wavevectors = np.stack(
                    np.meshgrid(
                        *[2 * np.pi * np.arange(n_points) / L for L in dimensions]
                    ),
                    axis=-1,
                ).reshape(-1, 3)
        self._wavenumbers = np.linalg.norm(self._wavevectors, axis=1)

        if q_max is not None:
            q_max = strip_unit(q_max, "Å^-1")[0]
            keep = self._wavenumbers <= q_max
            self._wavevectors = self._wavevectors[keep]
            self._wavenumbers = self._wavenumbers[keep]

        self._njit = lambda s: numba.njit(s, fastmath=True, parallel=parallel)
        self._delta_fourier_transform_sum = self._njit(
            structure.delta_fourier_transform_sum
        )
        self._ssf_trigonometric = self._njit(structure.ssf_trigonometric)
        self._psf_trigonometric = self._njit(structure.psf_trigonometric)
        self._inner = (
            accelerated.numba_inner_parallel if parallel else accelerated.numba_inner
        )

        self._form = form
        self._sort = sort
        self._unique = unique
        self._verbose = verbose

    def _prepare(self) -> None:

        # Preallocate array to store single-chain structure factors
        if not self._parallel:
            self.results.scsf = np.zeros((self._n_groups, len(self._wavenumbers)))

        # Determine the unique wavenumbers
        self.results.wavenumbers = np.unique(self._wavenumbers.round(11))

        # Store reference units
        self.results.units = Hash({"results.wavenumbers": ureg.angstrom**-1})

    def _single_frame(self) -> None:

        for ig, (ag, gr, M, N) in enumerate(
            zip(self._groups, self._groupings, self._n_chains, self._n_monomers)
        ):

            # Get entity positions in the current frame
            if self._internal and gr == "residues":
                positions = center_of_mass(ag, gr)
            else:
                positions = (
                    ag.positions
                    if gr == "atoms"
                    else center_of_mass(
                        positions=ag.positions.reshape(M, N, -1, 3),
                        masses=ag.masses.reshape(M, N, -1),
                    )
                )

            # Calculate single-chain structure factor contributions
            if self._form == "exp":
                for chain in positions.reshape((M, N, 3)):
                    rhos = self._delta_fourier_transform_sum(self._wavevectors, chain)
                    self.results.scsf[ig] += (rhos * rhos.conj()).real
            elif self._form == "trig":
                for chain in positions.reshape((M, N, 3)):
                    self.results.scsf[ig] += self._ssf_trigonometric(
                        self._inner(self._wavevectors, chain)
                    )

    def _conclude(self) -> None:

        # Normalize the single-chain structure factor by the number of
        # monomers and timesteps
        self.results.scsf /= self._n_chains * self._n_monomers * self.n_frames

        # Combine values sharing the same wavenumber, if desired
        if self._unique:
            self.results.scsf = np.hstack(
                [
                    self.results.scsf[:, np.isclose(q, self._wavenumbers)].mean(
                        axis=1, keepdims=True
                    )
                    for q in self.results.wavenumbers
                ]
            )

        # Sort the results by wavenumber, if desired
        if self._sort:
            order = np.argsort(self.results.wavenumbers)
            self.results.wavenumbers = self.results.wavenumbers[order]
            self.results.scsf = self.results.scsf[:, order]



r"""
Weibull distribution models
===========================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

The Weibull distribution is widely used in reliability and life (failure
rate) data analysis. This module provides the one-, two-, and
three-parameter Weibull distributions.

The three-parameter Weibull distribution is

.. math::

   y=ab(x-c)^{b-1}\exp{[-a(x-c)^b]}

where :math:`a` is the scale parameter, :math:`b` is the shape
parameter, and :math:`c` is the location parameter.

The two-parameter Weibull distribution

.. math::

   y=abx^{b-1}\exp{(-ax^b)}

has :math:`x-c` replaced with :math:`x`.

The one-parameter Weibull distribution has the shape parameter fixed,
so only the scale parameter is fitted.
"""

import numpy as np


def weibull(x: np.ndarray, a: float, b: float, c: float = 0) -> np.ndarray:
    r"""
    General three-parameter Weibull distribution.

    .. math::

       y=ab(x-c)^{b-1}\exp{[-a(x-c)^b]}

    Parameters
    ----------
    x : `numpy.ndarray`
        One-dimensional array containing :math:`x`-values.

    a : `float`
        Scale parameter :math:`a`.

    b : `float`
        Shape parameter :math:`b`. If specified to be a constant, the
        one-parameter Weibull distribution is used.

    c : `float`, keyword-only, default: :code:`0`
        Location parameter :math:`c`. If not specified as a parameter,
        the two-parameter Weibull distribution is used.

    Returns
    -------
    fit : `numpy.ndarray`
        Fitted :math:`y`-values.

    Examples
    --------
    Create a three-parameter Weibull distribution model for fitting.

    >>> model = lambda x, a, b, c: weibull(x, a, b, c)

    Create a two-parameter Weibull distribution model for fitting.

    >>> model = lambda x, a, b: weibull(x, a, b)

    Create a one-parameter Weibull distribution model for fitting, with
    :math:`b = 1`.

    >>> model = lambda x, a: weibull(x, a, 1)
    """

    return a * b * (x - c) ** (b - 1) * np.exp(-a * (x - c) ** b)



"""
Curve fitting models
====================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module provides a library of curve fitting models, such as
exponential, Fourier series, polynomial, and power law models, for use
with :func:`scipy.optimize.curve_fit`.
"""

from . import distribution, exponential, fourier, gaussian, polynomial, power

__all__ = ["distribution", "exponential", "fourier", "gaussian", "polynomial", "power"]
